"""
                Symbolic dynamics preprocessing (CINC2021)

get_sym_features - ready
get_Uspensky_coding - ready
maxmin_coding - ready
get_sigma_coding - ready
get_delta_coding - ready
get_delta_coding_binary - ready
get_ordinal_pattern - ready
perm_indices - ready

Version: 1.1 - CINC2021
Status: ready
2021/03/09
By Krivenko S.S.
"""
import numpy as np
from scipy.special import factorial

# -------------------------------- [ get_sym_features ] -------------------------------- #
def get_sym_features(ecg_features_lite):
    """
        Collect symbolic features

    Args:        
        ecg_features_lite: dict with data

    Returns:
        sym_features: dict with symbolic data
        single_element: single element value
    """
    sym_features = dict()    

    # Ordinal patterns
    base_name = 'op'
    wls = [3]
    tls = [1]
    for wl in wls:
        for tl in tls:            
            hist_array, single_element = get_ordinal_pattern(
                ecg_features_lite['times']['current_ibi'], wl, tl, len(ecg_features_lite['times']['current_ibi']))
            sym_features[base_name + '_' + str(wl) + '_' + str(tl)] = zeroing_indexes(hist_array, single_element)

    # Delta coding (binary)
    base_name = 'dcbin'
    wls = [3]
    tls = [1]    
    for wl in wls:
        for tl in tls:
            hist_array, single_element = get_delta_coding_binary(
                ecg_features_lite['times']['current_ibi'], wl, tl, len(ecg_features_lite['times']['current_ibi']))    
            sym_features[base_name + '_' + str(wl) + '_' + str(tl)] = zeroing_indexes(hist_array, single_element)

    # Delta coding
    base_name = 'dc'
    wls = [4]
    tls = [1]
    deltas = [5, 10, 20]
    for wl in wls:
        for tl in tls:
            for delta in deltas:
                hist_array, single_element = get_delta_coding(
                    ecg_features_lite['times']['current_ibi'], wl, tl, len(ecg_features_lite['times']['current_ibi']), delta)
                sym_features[base_name + '_' + str(wl) + '_' + str(tl) + '_' + str(delta)] = zeroing_indexes(hist_array, single_element)

    # Maxmin coding
    base_name = 'mmc'
    wls = [1]
    tls = [1, 2]
    for wl in wls:
        for tl in tls:
            hist_array, single_element = maxmin_coding(
                ecg_features_lite['times']['current_ibi'], wl, tl, len(ecg_features_lite['times']['current_ibi']))
            sym_features[base_name + '_' + str(wl) + '_' + str(tl)] = zeroing_indexes(hist_array, single_element)

    # Uspensky coding
    base_name = 'uc'
    wls = [1]
    tls = [1, 2]
    for wl in wls:
        for tl in tls:
            hist_array, single_element = get_Uspensky_coding(
                ecg_features_lite['times']['current_ibi'], ecg_features_lite['mags']['current_mags'], wl, tl, len(ecg_features_lite['times']['current_ibi']))
            sym_features[base_name + '_' + str(wl) + '_' + str(tl)] = zeroing_indexes(hist_array, single_element)

    return sym_features
# -------------------------------- [  ] -------------------------------- #

# -------------------------------- [ zeroing_indexes ] -------------------------------- #
def zeroing_indexes(input_array, se):
    """
        Set to zero elements <= 2*single_element

    Args:        
        input_array: input numpy array
        se: single element

    Returns:
        output_array: output numpy array        
    """
    output_array = input_array - 2*se
    neg_idx = np.nonzero(output_array < 0)
    output_array[neg_idx] = 0
    return output_array
# -------------------------------- [  ] -------------------------------- #    

# -------------------------------- [ get_Uspensky_coding ] -------------------------------- #
def get_Uspensky_coding(rr, mm, word_length, time_lag, rr_amount):
    """
        Get Uspensky coding scheme aka 6RTAL
        from "ТЕОРИЯ И ПРАКТИКА ДИАГНОСТИКИ ЗАБОЛЕВАНИЙ ВНУТРЕННИХ ОРГАНОВ
        МЕТОДОМ ИНФОРМАЦИОННОГО АНАЛИЗА ЭЛЕКТРОКАРДИОСИГНАЛОВ"

    Args:        
        rr: RR series
        mm: magnitudes corresponding to R-peaks in RR series
        word_length: word length (2, 3)
        time_lag: time lag (1,2)
        rr_amount: length of required RR-series

    Returns:
        patterns: output histogram
        single_element: single element weight
    """    
    power_base = 6
    bins_amount = np.power(power_base, word_length)
    target_rr = rr[:rr_amount:time_lag]
    target_mm = mm[:rr_amount:time_lag]
    target_rr_1 = np.roll(target_rr, -1)
    target_mm_1 = np.roll(target_mm, -1)
    target_atan = np.arctan(np.divide(target_mm, target_rr))
    target_atan_1 = np.arctan(np.divide(target_mm_1, target_rr_1))
    bool_seq = np.array(target_rr)
    seq0 = np.nonzero((target_mm < target_mm_1) & (target_rr < target_rr_1) & (target_atan < target_atan_1))
    seq1 = np.nonzero((target_mm >= target_mm_1) & (target_rr >= target_rr_1) & (target_atan < target_atan_1))
    seq2 = np.nonzero((target_mm < target_mm_1) & (target_rr >= target_rr_1) & (target_atan < target_atan_1))
    seq3 = np.nonzero((target_mm >= target_mm_1) & (target_rr < target_rr_1) & (target_atan >= target_atan_1))
    seq4 = np.nonzero((target_mm < target_mm_1) & (target_rr < target_rr_1) & (target_atan >= target_atan_1))
    seq5 = np.nonzero((target_mm >= target_mm_1) & (target_rr >= target_rr_1) & (target_atan >= target_atan_1))
    bool_seq[seq0] = 0
    bool_seq[seq1] = 1
    bool_seq[seq2] = 2
    bool_seq[seq3] = 3
    bool_seq[seq4] = 4
    bool_seq[seq5] = 5
    bool_seq = bool_seq.astype(int)
    M = len(bool_seq)
    bool_seq_idx = np.arange(0, M - word_length, 1)
    ready_seq = np.array([])
    for i in bool_seq_idx:
        local_seq = bool_seq[i:i + word_length]
        ready_seq = np.append(ready_seq, local_seq.dot(np.power(power_base, np.arange(local_seq.size)[::-1])))
    ready_seq = ready_seq.astype(int)    
    patterns = np.histogram(ready_seq, bins=bins_amount, range=(0, bins_amount))[0]
    N = len(ready_seq)
    single_element = np.divide(1, N)
    patterns = np.divide(patterns, N)
    return patterns, single_element
# -------------------------------- [  ] -------------------------------- #

# -------------------------------- [ maxmin_coding ] -------------------------------- #
def maxmin_coding(rr, word_length, time_lag, rr_amount):
    """
        Get maxmin coding scheme aka Wessel scheme
        from "Quantifying heart rate dynamics using different approaches of symbolic dynamics"

    Args:        
        rr: RR series
        word_length: word length (2, 3)
        time_lag: time lag (1,2)
        rr_amount: length of required RR-series

    Returns:
        patterns: output histogram
        single_element: single element weight
    """        
    power_base = 6
    bins_amount = np.power(power_base, word_length)    
    target_rr = rr[:rr_amount:time_lag]
    bool_seq = np.array(target_rr)
    minx = np.amin(bool_seq)
    maxx = np.amax(bool_seq)
    l = (maxx - minx) / power_base
    seq0 = np.nonzero((target_rr >= minx) & (target_rr < 1*l+minx))
    seq1 = np.nonzero((target_rr >= 1*l+minx) & (target_rr < 2*l+minx))
    seq2 = np.nonzero((target_rr >= 2*l+minx) & (target_rr < 3*l+minx))
    seq3 = np.nonzero((target_rr >= 3*l+minx) & (target_rr < 4*l+minx))
    seq4 = np.nonzero((target_rr >= 4*l+minx) & (target_rr < 5*l+minx))
    seq5 = np.nonzero((target_rr >= 5*l+minx) & (target_rr <= maxx))
    bool_seq[seq0] = 0
    bool_seq[seq1] = 1
    bool_seq[seq2] = 2
    bool_seq[seq3] = 3
    bool_seq[seq4] = 4
    bool_seq[seq5] = 5
    bool_seq = bool_seq.astype(int)
    M = len(bool_seq)
    bool_seq_idx = np.arange(0, M - word_length, 1)
    ready_seq = np.array([])
    for i in bool_seq_idx:
        local_seq = bool_seq[i:i + word_length]
        ready_seq = np.append(ready_seq, local_seq.dot(np.power(power_base, np.arange(local_seq.size)[::-1])))
    ready_seq = ready_seq.astype(int)
    patterns = np.histogram(ready_seq, bins=bins_amount, range=(0, bins_amount))[0]
    N = len(ready_seq)
    single_element = np.divide(1, N)
    patterns = np.divide(patterns, N)
    return patterns, single_element
# -------------------------------- [  ] -------------------------------- #  

# -------------------------------- [ get_sigma_coding ] -------------------------------- #
def get_sigma_coding(rr, word_length, time_lag, rr_amount):
    """
        Get sigma coding scheme aka Wessel scheme
        from "Quantifying heart rate dynamics using different approaches of symbolic dynamics"

    Args:        
        rr: RR series
        word_length: word length (3)
        time_lag: time lag (1,2)
        rr_amount: length of required RR-series

    Returns:
        patterns: output histogram
        single_element: single element weight
    """    
    alpha = 0.05
    power_base = 4
    bins_amount = np.power(power_base, word_length)    
    target_rr = rr[:rr_amount:time_lag]
    bool_seq = np.array(target_rr)
    mu = np.mean(target_rr)
    seq0 = np.nonzero(np.logical_and(target_rr <= (1 + alpha)*mu, target_rr > mu))
    seq1 = np.nonzero(target_rr > (1 + alpha) * mu)
    seq2 = np.nonzero(np.logical_and(target_rr <= mu, target_rr > (1 - alpha)*mu))
    seq3 = np.nonzero(np.logical_and(target_rr <= (1 - alpha)*mu, target_rr > 0))
    bool_seq[seq0] = 0
    bool_seq[seq1] = 1
    bool_seq[seq2] = 2
    bool_seq[seq3] = 3
    bool_seq = bool_seq.astype(int)
    M = len(bool_seq)
    bool_seq_idx = np.arange(0, M - word_length, 1)
    ready_seq = np.array([])
    for i in bool_seq_idx:
        local_seq = bool_seq[i:i + word_length]
        ready_seq = np.append(ready_seq, local_seq.dot(np.power(power_base, np.arange(local_seq.size)[::-1])))
    ready_seq = ready_seq.astype(int)
    patterns = np.histogram(ready_seq, bins=bins_amount, range=(0, bins_amount))[0]
    N = len(ready_seq)
    single_element = np.divide(1, N)
    patterns = np.divide(patterns, N)
    return patterns, single_element
# -------------------------------- [  ] -------------------------------- #  

# -------------------------------- [ get_delta_coding_0 ] -------------------------------- #
def get_delta_coding(rr, word_length, time_lag, rr_amount, delta=5):
    """
        Get delta coding scheme aka PolVar5 (delta = 5 ms)
        from "Quantifying heart rate dynamics using different approaches of symbolic dynamics"

    Args:        
        rr: RR series
        word_length: word length (3,4,5,6,7,8)
        time_lag: time lag (1,2,3,4,5,6,7,8)
        rr_amount: length of required RR-series
        delta: delta (5,10,20)

    Returns:
        patterns: output histogram
        single_element: single element weight
    """    
    delta = delta / 1000
    power_base = 2
    bins_amount = np.power(power_base, word_length)    
    target_rr = rr[:rr_amount:time_lag]
    diffs = np.absolute(np.diff(target_rr))
    bool_seq = (diffs >= delta).astype(int)
    M = len(bool_seq)
    bool_seq_idx = np.arange(0, M - word_length, 1)
    ready_seq = np.array([])
    for i in bool_seq_idx:
        local_seq = bool_seq[i:i + word_length]
        ready_seq = np.append(ready_seq, local_seq.dot(np.power(power_base, np.arange(local_seq.size)[::-1])))
    ready_seq = ready_seq.astype(int)
    patterns = np.histogram(ready_seq, bins=bins_amount, range=(0, bins_amount))[0]
    N = len(ready_seq)
    single_element = np.divide(1, N)
    patterns = np.divide(patterns, N)
    return patterns, single_element
# -------------------------------- [  ] -------------------------------- #  

# -------------------------------- [ get_delta_coding_binary ] -------------------------------- #
def get_delta_coding_binary(rr, word_length, time_lag, rr_amount):
    """
        Get delta coding scheme (binary) aka Cysarz
        from "Quantifying heart rate dynamics using different approaches of symbolic dynamics"

    Args:        
        rr: RR series
        word_length: word length (3,4,5,6,7,8)
        time_lag: time lag (1,2,3,4,5,6,7,8)
        rr_amount: length of required RR-series

    Returns:
        patterns: output histogram
        single_element: single element weight
    """    
    power_base = 2
    bins_amount = np.power(power_base, word_length)
    target_rr = rr[:rr_amount:time_lag]
    diffs = np.diff(target_rr)
    bool_seq = (diffs <= 0).astype(int)
    M = len(bool_seq)
    bool_seq_idx = np.arange(0, M - word_length, 1)
    ready_seq = np.array([])
    for i in bool_seq_idx:
        local_seq = bool_seq[i:i + word_length]
        ready_seq = np.append(ready_seq, local_seq.dot(np.power(power_base, np.arange(local_seq.size)[::-1])))
    ready_seq = ready_seq.astype(int)
    patterns = np.histogram(ready_seq, bins=bins_amount, range=(0, bins_amount))[0]    
    N = len(ready_seq)
    single_element = np.divide(1, N)
    patterns = np.divide(patterns, N)
    return patterns, single_element
# -------------------------------- [  ] -------------------------------- #    

# -------------------------------- [ get_ordinal_pattern ] -------------------------------- #
def get_ordinal_pattern(rr, word_length, time_lag, rr_amount):
    """
        Get norm Ordinal Pattern
        from "Classifying Cardiac Biosignals using Ordinal Pattern Statistics and Symbolic Dynamics"

    Args:        
        rr: RR series
        word_length: word length (3,4,5)
        time_lag: time lag (1,2,3,4,5,6,7,8)
        rr_amount: length of required RR-series

    Returns:
        ordinal_pattern: output histogram
        single_element: single element weight
    """    
    target_rr = rr[:rr_amount:time_lag]
    indcs = perm_indices(target_rr, word_length, time_lag) - 1
    N = len(indcs)
    M = int(factorial(word_length))
    single_element = np.divide(1, N)
    #ordinal_pattern = np.histogram(indcs, bins=M, range=(1, M + 1))[0]    
    ordinal_pattern = np.histogram(indcs, bins=M, range=(0, M))[0]
    ordinal_pattern = np.divide(ordinal_pattern, N)  # normalization of ordinal pattern
    return ordinal_pattern, single_element
# -------------------------------- [  ] -------------------------------- #    

# -------------------------------- [ perm_indices ] -------------------------------- #
def perm_indices(ts, wl=4, lag=1):
    """
        Computing permutation indices
        from "Classifying Cardiac Biosignals using Ordinal Pattern Statistics and Symbolic Dynamics"

    Args:        
        ts: time series
        wl: word length
        lag: time lag

    Returns:
        indcs: permutation indices
    """
    m = len(ts) - (wl - 1) * lag
    indcs = np.zeros(m, dtype=int)
    for i in range(1, wl):
        st = ts[(i - 1)*lag:m + ((i - 1)*lag)]
        for j in range(i, wl):
            indcs += st > ts[j*lag:m + j*lag]
        indcs *= wl - i
    return indcs + 1
# -------------------------------- [  ] -------------------------------- #