"""
                Helper functions for CINC2021

xgboost_prepare_balanced - 
balance_dataset_optim - ready
balance_dataset - ready
xgboost_prepare - 
CreateBalancedSampleWeights - ready
get_features_as_image - ready
get_features_mfcc - ready
get_features_neurokit - ready
plot_2ecg - ready
plot_ecg - ready
plotting_ECG_points - ready
create_train_test_folders_all_class - ready
create_train_test_folders_any_class - ready
create_train_test_folders - ready
reduce_mem_usage - ready
resample - ready
repeat_until_n - ready
ecg_preprocess - ready
peakDetectCorrect - ready
detect_bad_frames - ready
delete_bad_frames - ready
crop_or_pad - ready
mono_to_color - ready

Version: 1.0
Status: 
2021/03/02
By Krivenko S.S.
"""
import numpy as np
from numpy import inf, nan
import pandas as pd
import time
import ntpath
from shutil import copyfile
from scipy.special import factorial
from scipy import signal as sig
from helper_code import *
import matplotlib.pyplot as plt
import neurokit2 as nk
from neurokit2.signal import signal_filter
from sklearn.ensemble import RandomForestClassifier
from xgboost import XGBClassifier
import librosa
import librosa.display as lbd
from sklearn.model_selection import train_test_split, StratifiedKFold
from sklearn.metrics import roc_auc_score, label_ranking_average_precision_score, precision_recall_curve, auc, confusion_matrix
from wfdb import processing
from pyts.image import GramianAngularField, MarkovTransitionField, RecurrencePlot
from pyts.multivariate.image import JointRecurrencePlot
from pyts.datasets import load_gunpoint, load_basic_motions
#from mpl_toolkits.axes_grid1 import ImageGrid
import cv2

import ecg_symbolic as ecg_sym

# Define the Challenge lead sets. These variables are not required. You can change or remove them.
twelve_leads = ('I', 'II', 'III', 'aVR', 'aVL', 'aVF', 'V1', 'V2', 'V3', 'V4', 'V5', 'V6')
six_leads = ('I', 'II', 'III', 'aVR', 'aVL', 'aVF')
four_leads = ('I', 'II', 'III', 'V2')
three_leads = ('I', 'II', 'V2')
two_leads = ('I', 'II')

# -------------------------------- [ rf_prepare_balanced ] -------------------------------- #
def rf_prepare_balanced(features, labels, optimize=False, random_state=42):
    """
    Prepare random forest classifier, balanced classes

    Args:
        features: input data, 2D numpy array
        labels: targets, 2D one-hot numpy array
        optimize: optimization flag
        random_state: random state

    Returns:
        model: ready-to-use XGBoost classifier

    """    
    #labels = transform_ohe_labels(labels)
    n_splits = 5
    skf = StratifiedKFold(n_splits=n_splits, shuffle=True, random_state=random_state)
    models = dict()
    scores = dict()
    sensitivity = dict()
    specificity = dict()

    for i in range(0, labels.shape[1]):
        print('Model for {0}:'.format(i))
        Xfull, yfull = balance_dataset_optim(features, labels, i)        
        train_index, val_index = list(skf.split(Xfull, yfull))[0]
        model = RandomForestClassifier(
                                   random_state=random_state,
                                  )
        model.fit(X=Xfull[train_index, :], 
                        y=yfull[train_index], 
                        )
            
        yvalid = model.predict_proba(Xfull[val_index, :])[:,1]
        ypred = model.predict(Xfull[val_index, :])
        precision, recall, thresholds = precision_recall_curve(yfull[val_index], yvalid)
        models[str(i)] = model
        scores[str(i)] = auc(recall, precision)
        tn, fp, fn, tp = confusion_matrix(yfull[val_index], ypred).ravel()
        sensitivity[str(i)] = tp / (tp + fn)
        specificity[str(i)] = tn / (tn + fp)
        print('Model #{0}, TPR = {1}, TNR = {2}'.format(i, sensitivity[str(i)], specificity[str(i)]))

    return models, scores, sensitivity, specificity
# -------------------------------- [ ] -------------------------------- #

# -------------------------------- [ xgboost_prepare_balanced ] -------------------------------- #
def xgboost_prepare_balanced(features, labels, optimize=False, random_state=42):
    """
    Prepare XGBoost classifier (XGBoost x num_classes), balanced classes

    Args:
        features: input data, 2D numpy array
        labels: targets, 2D one-hot numpy array
        optimize: optimization flag
        random_state: random state

    Returns:
        model: ready-to-use XGBoost classifier

    """    
    #labels = transform_ohe_labels(labels)
    n_splits = 5
    skf = StratifiedKFold(n_splits=n_splits, shuffle=True, random_state=random_state)
    models = dict()
    scores = dict()
    sensitivity = dict()
    specificity = dict()

    for i in range(0, labels.shape[1]):
        print('Model for {0}:'.format(i))
        Xfull, yfull = balance_dataset_optim(features, labels, i)        
        train_index, val_index = list(skf.split(Xfull, yfull))[0]
        model = XGBClassifier(n_estimators=500,
                                   max_depth=5,
                                   learning_rate=0.01,
                                   verbosity=0,
                                   objective='binary:logistic',
                                   #subsample=0.95,
                                   #colsample_bytree=0.95,
                                   random_state=random_state,
                                   #tree_method='gpu_hist',
                                   #predictor='gpu_predictor',
                                   use_label_encoder=False,
                                   scale_pos_weight = 1
                                  )
        model.fit(X=Xfull[train_index, :], 
                        y=yfull[train_index], 
                        eval_set=[(Xfull[val_index, :], yfull[val_index])], 
                        eval_metric='aucpr', 
                        early_stopping_rounds=100, 
                        verbose=100)
            
        yvalid = model.predict_proba(Xfull[val_index, :])[:,1]
        ypred = model.predict(Xfull[val_index, :])
        precision, recall, thresholds = precision_recall_curve(yfull[val_index], yvalid)
        models[str(i)] = model
        scores[str(i)] = auc(recall, precision)
        tn, fp, fn, tp = confusion_matrix(yfull[val_index], ypred).ravel()
        sensitivity[str(i)] = tp / (tp + fn)
        specificity[str(i)] = tn / (tn + fp)
        print('Model #{0}, TPR = {1}, TNR = {2}'.format(i, sensitivity[str(i)], specificity[str(i)]))

    return models, scores, sensitivity, specificity
# -------------------------------- [ ] -------------------------------- #

# -------------------------------- [ balance_dataset_optim ] -------------------------------- #
def balance_dataset_optim(features, labels, class_number):
    """
    Balance dataset for binary classification in multiclass multilabel task, optimized

    Args:
        ecg: ecg signal in mV

    Returns:

    """
    target = labels[:, class_number].astype(int)
    pos_idx = target == 1
    size_of_dataset = 2 * np.sum(pos_idx)
    Xfull = features[pos_idx, :]
    yfull = np.ones(np.sum(pos_idx)).astype(int)
    neg_idx = np.zeros([np.sum(pos_idx)])
    neg_count = 0
    next_class = 0
    classes_amount = 27
    start_time = time.time()
    next_time = 0
    while neg_count < size_of_dataset / 2 and next_time < 300:
        next_time = time.time() - start_time
        for j in range(features.shape[0]):
            label = labels[j, :]
            if next_class>=classes_amount:
                next_class = 0
            elif next_class==class_number:
                next_class += 1
            elif label[next_class] == 1 and label[class_number] == 0 and neg_count < size_of_dataset / 2:
                if j not in neg_idx:
                    neg_idx[neg_count] = j
                    neg_count += 1
                    next_class += 1
            if neg_count >= size_of_dataset / 2:
                break

    if neg_count < size_of_dataset / 2 - 1:
        for j in range(features.shape[0]):            
            label = labels[j, :]
            if sum(label) == 0 and neg_count < size_of_dataset / 2:
                if j not in neg_idx:
                    neg_idx[neg_count] = j
                    neg_count += 1
            elif label[class_number] == 0 and neg_count < size_of_dataset / 2:
                if j not in neg_idx:
                    neg_idx[neg_count] = j
                    neg_count += 1
            if neg_count >= size_of_dataset / 2:
                break
    
    Xfull = np.vstack((Xfull, features[neg_idx.astype(int), :]))
    yfull = np.append(yfull, np.zeros(np.sum(pos_idx)).astype(int))
    
    print('Unique negative idx amount: {0}'.format(np.unique(neg_idx).size))
    print('Positive amount: {0}, Negative amount: {1}'.format(np.sum(pos_idx), Xfull.shape[0]-np.sum(pos_idx)))
    return Xfull, yfull
# -------------------------------- [ ] -------------------------------- #     

# -------------------------------- [ balance_dataset ] -------------------------------- #
def balance_dataset(features, labels, class_number):
    """
    Balance dataset for binary classification in multiclass multilabel task

    Args:
        ecg: ecg signal in mV

    Returns:

    """
    target = labels[:, class_number].astype(int)
    pos_idx = target == 1
    size_of_dataset = 2 * np.sum(pos_idx)
    Xfull = features[pos_idx, :]
    yfull = np.ones(np.sum(pos_idx)).astype(int)
    next_class = 0
    classes_amount = 27
    start_time = time.time()
    next_time = 0
    while len(yfull) < size_of_dataset and next_time < 60:
        next_time = time.time() - start_time
        for j in range(features.shape[0]):
            feature = features[j, :]
            label = labels[j, :]
            if next_class>=classes_amount:
                next_class = 0
            elif next_class==class_number:
                next_class += 1
            elif label[next_class] == 1 and label[class_number] == 0 and len(yfull) < size_of_dataset:
                if not any(np.equal(Xfull, feature).all(1)):
                    Xfull = np.vstack((Xfull, feature))
                    yfull = np.append(yfull, 0)
                    next_class += 1
            if len(yfull) >= size_of_dataset:
                break

    for j in range(features.shape[0]):
            feature = features[j, :]
            label = labels[j, :]
            if sum(label) == 0 and len(yfull) < size_of_dataset:
                if not any(np.equal(Xfull, feature).all(1)):
                    Xfull = np.vstack((Xfull, feature))
                    yfull = np.append(yfull, 0)
            elif label[class_number] == 0 and len(yfull) < size_of_dataset:
                if not any(np.equal(Xfull, feature).all(1)):
                    Xfull = np.vstack((Xfull, feature))
                    yfull = np.append(yfull, 0)
            if len(yfull) >= size_of_dataset:
                break
    
    print('Positive amount: {0}, Negative amount: {1}'.format(np.sum(pos_idx), Xfull.shape[0]-np.sum(pos_idx)))
    return Xfull, yfull
# -------------------------------- [ ] -------------------------------- #     

# -------------------------------- [ xgboost_prepare ] -------------------------------- #
def xgboost_prepare(features, labels, optimize=False, random_state=42):
    """
    Prepare XGBoost classifier (XGBoost x num_classes) 

    Args:
        features: input data, 2D numpy array
        labels: targets, 2D one-hot numpy array
        optimize: optimization flag
        random_state: random state

    Returns:
        model: ready-to-use XGBoost classifier

    """    
    #labels = transform_ohe_labels(labels)
    n_splits = 5
    skf = StratifiedKFold(n_splits=n_splits, shuffle=True, random_state=random_state)
    models = dict()
    scores = dict()
    sensitivity = dict()
    specificity = dict()

    for i in range(0, labels.shape[1]):
        print('Model for {0}:'.format(i))
        starttime = time.time()
        target = labels[:, i].astype(int)
        train_index, val_index = list(skf.split(features, target))[0]
        #largest_class_weight_coef = np.sum(target[train_index]==0) / len(target[train_index])
        #sample_weights = CreateBalancedSampleWeights(target[train_index], largest_class_weight_coef)
        #for fold_id, (train_index, val_index) in enumerate(skf.split(features, target)):        
        model = XGBClassifier(n_estimators=500,
                                   max_depth=10,
                                   learning_rate=0.01,
                                   verbosity=0,
                                   objective='binary:logistic',
                                   subsample=0.95,
                                   colsample_bytree=0.95,
                                   random_state=random_state,
                                   #tree_method='gpu_hist',
                                   #predictor='gpu_predictor',
                                   use_label_encoder=False,
                                   #weight = sample_weights,
                                   scale_pos_weight = np.sum(target==0) / np.sum(target==1),
                                   #scale_pos_weight = 1,
                                  )
        model.fit(X=features[train_index, :], 
                        y=target[train_index], 
                        eval_set=[(features[val_index, :], target[val_index])], 
                        eval_metric='aucpr', 
                        early_stopping_rounds=100, 
                        verbose=100)
            
        yvalid = model.predict_proba(features[val_index, :])[:,1]
        ypred = model.predict(features[val_index, :])
        precision, recall, thresholds = precision_recall_curve(target[val_index], yvalid)
        models[str(i)] = model
        scores[str(i)] = auc(recall, precision)
        tn, fp, fn, tp = confusion_matrix(target[val_index], ypred).ravel()
        sensitivity[str(i)] = tp / (tp + fn)
        specificity[str(i)] = tn / (tn + fp)
        print('Model #{0}, TPR = {1}, TNR = {2}'.format(i, sensitivity[str(i)], specificity[str(i)]))

    return models, scores, sensitivity, specificity
# -------------------------------- [ ] -------------------------------- #

# -------------------------------- [ CreateBalancedSampleWeights ] -------------------------------- #
def CreateBalancedSampleWeights(y_train, largest_class_weight_coef):
    classes = np.unique(y_train, axis = 0)
    classes.sort()
    class_samples = np.bincount(y_train)
    total_samples = class_samples.sum()
    n_classes = len(class_samples)
    weights = total_samples / (n_classes * class_samples * 1.0)
    class_weight_dict = {key : value for (key, value) in zip(classes, weights)}
    class_weight_dict[classes[1]] = class_weight_dict[classes[1]] * largest_class_weight_coef
    sample_weights = [class_weight_dict[y] for y in y_train]
    return sample_weights
# -------------------------------- [ ] -------------------------------- #    

# -------------------------------- [ get_features_as_image ] -------------------------------- #
# Extract features from the header and recording.
def get_features_as_image(header, recording, leads, image_size=(224,224), image_type='mel'):
    # Extract age.
    age = get_age(header)
    if age is None:
        age = float('nan')

    # Extract sex. Encode as 0 for female, 1 for male, and NaN for other.
    sex = get_sex(header)
    if sex in ('Female', 'female', 'F', 'f'):
        sex = 0
    elif sex in ('Male', 'male', 'M', 'm'):
        sex = 1
    else:
        sex = float('nan')

    # Extract fs.
    fs = int(get_frequency(header))
    if fs is None:
        fs = 500    

    # Reorder/reselect leads in recordings.
    available_leads = get_leads(header)
    indices = list()
    for lead in leads:
        i = available_leads.index(lead)
        indices.append(i)
    recording = recording[indices, :].astype(np.float32)

    # Pre-process recordings.
    adc_gains = get_adc_gains(header, leads)
    baselines = get_baselines(header, leads)
    num_leads = len(leads)
    for i in range(num_leads):
        recording[i, :] = (recording[i, :] - baselines[i]) / adc_gains[i]

    #plot_ecg(recording[0, :])
    
    # Preprocess ECG
    #rms = np.zeros(num_leads, dtype=np.float32)
    f500 = 500  # 500 Hz
    sig_len_sec = 10  # 10 seconds

    sym_to_append = np.array([])
    for i in range(num_leads):        
        x = recording[i, :]        
        signal = resample(x, fs, f500)
        signal = ecg_preprocess(signal, f500)  # preprocessing
        #plot_2ecg(x, signal)
        #signal = repeat_until_n(signal, sig_len_sec * f500)  # 10 sec segment
        signal = crop_or_pad(signal, sig_len_sec * f500, is_train=True, start=1)
        signal[np.isnan(signal)] = 0
        sym_to_append = np.append(sym_to_append, signal)
        
    if image_type == 'mel':
        #plot_2ecg(sym_to_append, sym_to_append)        

        mel_spec = librosa.feature.melspectrogram(sym_to_append, n_mels=128, n_fft=2048, hop_length=512//40, sr=f500, power=1.2)
        #mel_spec = librosa.power_to_db(mel_spec).astype(np.float32)
        mel_spec = librosa.amplitude_to_db(mel_spec, ref = np.max)

        #mel_spec = np.abs(librosa.stft(sym_to_append, n_fft = 2048, hop_length = 512//40))
        #mel_spec = librosa.amplitude_to_db(mel_spec, ref = np.max)
        
        image = mono_to_color(mel_spec)

    elif image_type == 'chroma':        
        mel_spec = librosa.feature.chroma_stft(sym_to_append, sr=f500)
        image = mono_to_color(mel_spec) 
    elif image_type == 'gasf':
        gasf = GramianAngularField(image_size=image_size[0], method='summation')
        X_gasf = gasf.fit_transform(sym_to_append.reshape(1, -1))
        image = mono_to_color(X_gasf[0])
    elif image_type == 'gadf':
        gadf = GramianAngularField(image_size=image_size[0], method='difference')
        X_gadf = gadf.fit_transform(sym_to_append.reshape(1, -1))
        image = mono_to_color(X_gadf[0])
    elif image_type == 'mtf':
        mtf = MarkovTransitionField(image_size=image_size[0])
        X_mtf = mtf.fit_transform(sym_to_append.reshape(1, -1))
        image = mono_to_color(X_mtf[0])
        #lbd.specshow(image)
        #plt.show()
    elif image_type == 'rp':  # comp problem!
        rp = RecurrencePlot(threshold='point', percentage=20)
        X_rp = rp.fit_transform(sym_to_append.reshape(1, -1))
        image = mono_to_color(X_rp[0])        
        #lbd.specshow(image)
        #plt.show()
    elif image_type == 'jrp':
        sym_to_append = np.array([])
        for i in range(num_leads):        
            x = recording[i, :]        
            signal = resample(x, fs, f500)
            signal = ecg_preprocess(signal, f500)  # preprocessing
            #plot_2ecg(x, signal)
            #signal = repeat_until_n(signal, sig_len_sec * f500)  # 10 sec segment
            signal = crop_or_pad(signal, sig_len_sec * f500, is_train=True, start=1)
            signal[np.isnan(signal)] = 0
            if i == 0:
                sym_to_append = signal
            else:
                sym_to_append = np.vstack((sym_to_append, signal))
        
        jrp = JointRecurrencePlot(threshold='point', percentage=50)       
        X_jrp = jrp.fit_transform(np.reshape(sym_to_append, (1, sym_to_append.shape[0], sym_to_append.shape[1])))
        image = mono_to_color(X_jrp[0])
        #lbd.specshow(image)
        #plt.show()    
        
    image = cv2.resize(image, dsize=image_size, interpolation=cv2.INTER_AREA)
    if 1:
        #plot_ecg(sym_to_append)
        print(get_labels(header))
        lbd.specshow(image)
        plt.show()

    return image
# -------------------------------- [ ] -------------------------------- #

# -------------------------------- [ load_row ] -------------------------------- #
def load_row(row):
    header = load_header(row.header_files)
    recording = load_recording(row.recording_files)
    #t1 = time.time()
    if row.leads_amount == 12:
        local_data = get_features_neurokit_mfcc2(header, recording, twelve_leads)
    if row.leads_amount == 6:
        local_data = get_features_neurokit_mfcc2(header, recording, six_leads)
    if row.leads_amount == 4:
        local_data = get_features_neurokit_mfcc2(header, recording, four_leads)
    if row.leads_amount == 3:
        local_data = get_features_neurokit_mfcc2(header, recording, three_leads)
    if row.leads_amount == 2:
        local_data = get_features_neurokit_mfcc2(header, recording, three_leads)    

    print(header[:7])    
    #print('Time spent: {0}'.format(time.time() - t1))
    
    if row.is_train:
        mapping_scoring = pd.read_csv('dx_mapping_scored.csv')
        classes = list(mapping_scoring['SNOMED CT Code'].values.astype(str))
        #classes = list(['270492004', '164889003', '164890007'])
        if all(is_integer(x) for x in classes):
            classes = sorted(classes, key=lambda x: int(x)) # Sort classes numerically if numbers.
        else:
            classes = sorted(classes) # Sort classes alphanumerically otherwise.
        num_classes = len(classes)

        local_label = np.zeros((num_classes))
        current_labels = get_labels(header)
        for label in current_labels:
            if label in classes:
                j = classes.index(label)
                local_label[j] = 1
    else:
        local_label = None

    return local_data, local_label
# -------------------------------- [ ] -------------------------------- #

# -------------------------------- [ get_features_neurokit_mfcc2 ] -------------------------------- #
# Extract features from the header and recording.
def get_features_neurokit_mfcc2(header, recording, leads):
    # Extract age.
    age = get_age(header)
    if age is None:
        age = float('nan')

    # Extract sex. Encode as 0 for female, 1 for male, and NaN for other.
    sex = get_sex(header)
    if sex in ('Female', 'female', 'F', 'f'):
        sex = 0
    elif sex in ('Male', 'male', 'M', 'm'):
        sex = 1
    else:
        sex = float('nan')

    # Extract fs.
    fs = int(get_frequency(header))
    if fs is None:
        fs = 500

    # Reorder/reselect leads in recordings.
    available_leads = get_leads(header)
    indices = list()
    for lead in leads:
        i = available_leads.index(lead)
        indices.append(i)
    recording = recording[indices, :].astype(float)

    # Pre-process recordings.
    adc_gains = get_adc_gains(header, leads)
    baselines = get_baselines(header, leads)
    num_leads = len(leads)
    for i in range(num_leads):
        recording[i, :] = (recording[i, :] - baselines[i]) / adc_gains[i]

    #plot_ecg(recording[0, :])
    
    # Preprocess ECG
    #rms = np.zeros(num_leads, dtype=np.float32)
    f500 = 500  # 500 Hz
    f125 = 125  # 125 Hz
    sig_len_sec = 10  # 10 seconds
    sym_to_append = np.array([])
    for i in range(num_leads):        
        x = recording[i, :]        
        signal125 = resample(x, fs, f125)
        signal500 = resample(x, fs, f500)
        signal125 = ecg_preprocess(signal125, f125)  # preprocessing
        signal500 = ecg_preprocess(signal500, f500)  # preprocessing
        #plot_2ecg(x, signal)
        signal125 = repeat_until_n(signal125, sig_len_sec * f125)  # 10 sec segment
        signal125[np.isnan(signal125)] = 0
        signal500 = repeat_until_n(signal500, sig_len_sec * f500)  # 10 sec segment
        signal500[np.isnan(signal500)] = 0

        r_peaks = processing.xqrs_detect(sig=signal125, fs=f125, verbose=False, learn=True)
        if len(r_peaks) < 6:
            r_peaks = np.sort(np.random.choice(range(sig_len_sec * f125 - 1), sig_len_sec, replace=False))
        r_peaks = peakDetectCorrect(signal125, f125, r_peaks)
        r_mags = signal125[r_peaks]
        #plotting_ECG_points(signal125, r_peaks)

        try:
            hrv_indices = nk.hrv(r_peaks, sampling_rate=f125, show=False)
            hrv_values = np.squeeze(hrv_indices.values)
        except:
            hrv_values = np.zeros((57))
        
        if len(hrv_values) != 57:
            hrv_values = np.zeros((57))
        sym_to_append = np.append(sym_to_append, hrv_values)

        #bio_df, bio_info = nk.bio_process(ecg=signal, sampling_rate=f500)        
        #results = nk.bio_analyze(bio_df, sampling_rate=f500)
        #r_peaks = bio_df['ECG_R_Peaks'][bio_df['ECG_R_Peaks'] !=0].index.values.astype(int)
        #r_mags = bio_df['ECG_Clean'][r_peaks].values
        #plotting_ECG_points(bio_df['ECG_Clean'], r_peaks)       
 
        # Get sym features
        sym_values = np.array([])
        ecg_features_lite = dict()
        ecg_features_lite['times'] = dict()
        ecg_features_lite['times']['current_ibi'] = np.diff(r_peaks) / f125
        ecg_features_lite['mags'] = dict()
        ecg_features_lite['mags']['current_mags'] = r_mags[:-1]
        sym_features = ecg_sym.get_sym_features(ecg_features_lite)
        data_to_append = {}        
        cols = list(sym_features.keys())        
        for col in cols:
            chunk = sym_features[col]
            for i, value in enumerate(chunk):
                col_name = 'sym_' + col + '__' + str(i)
                data_to_append[col_name] = value
                sym_values = np.append(sym_values, value)

        if len(sym_values) != 86:
            sym_values = np.zeros((86))
        sym_to_append = np.append(sym_to_append, sym_values)

        # Get melspec
        mel_spec = librosa.feature.melspectrogram(signal500, n_mels=64, n_fft=2048, hop_length=512, sr=f500, power=2)
        if len(mel_spec.ravel()) != 640:
            mel_spec = np.zeros((64, 10))        
        sym_to_append = np.append(sym_to_append, mel_spec.ravel())

    # Check for nan and +-inf        
    sym_to_append[np.isinf(sym_to_append)] = 0
    sym_to_append[np.isneginf(sym_to_append)] = 0
    sym_to_append[np.isnan(sym_to_append)] = 0

    sym_to_append = np.append(sym_to_append, age)
    sym_to_append = np.append(sym_to_append, sex)
    return sym_to_append
# -------------------------------- [ ] -------------------------------- #

# -------------------------------- [ get_features_neurokit_mfcc ] -------------------------------- #
# Extract features from the header and recording.
def get_features_neurokit_mfcc(header, recording, leads):
    # Extract age.
    age = get_age(header)
    if age is None:
        age = float('nan')

    # Extract sex. Encode as 0 for female, 1 for male, and NaN for other.
    sex = get_sex(header)
    if sex in ('Female', 'female', 'F', 'f'):
        sex = 0
    elif sex in ('Male', 'male', 'M', 'm'):
        sex = 1
    else:
        sex = float('nan')

    # Extract fs.
    fs = int(get_frequency(header))
    if fs is None:
        fs = 500    

    # Reorder/reselect leads in recordings.
    available_leads = get_leads(header)
    indices = list()
    for lead in leads:
        i = available_leads.index(lead)
        indices.append(i)
    recording = recording[indices, :].astype(float)

    # Pre-process recordings.
    adc_gains = get_adc_gains(header, leads)
    baselines = get_baselines(header, leads)
    num_leads = len(leads)
    for i in range(num_leads):
        recording[i, :] = (recording[i, :] - baselines[i]) / adc_gains[i]

    #plot_ecg(recording[0, :])
    
    # Preprocess ECG
    #rms = np.zeros(num_leads, dtype=np.float32)
    f500 = 500  # 500 Hz
    sig_len_sec = 10  # 10 seconds
    sym_to_append = np.array([])
    for i in range(num_leads):        
        x = recording[i, :]        
        signal = resample(x, fs, f500)        
        signal = ecg_preprocess(signal, f500)  # preprocessing
        #plot_2ecg(x, signal)
        signal = repeat_until_n(signal, sig_len_sec * f500)  # 10 sec segment
        signal[np.isnan(signal)] = 0
        r_peaks = processing.xqrs_detect(sig=signal, fs=f500, verbose=False, learn=True)
        if len(r_peaks) < 6:            
            r_peaks = np.sort(np.random.choice(range(sig_len_sec * f500 - 1), sig_len_sec, replace=False))
        r_peaks = peakDetectCorrect(signal, f500, r_peaks)
        r_mags = signal[r_peaks]
        #plotting_ECG_points(signal, r_peaks)

        try:
            hrv_indices = nk.hrv(r_peaks, sampling_rate=f500, show=False)
            hrv_values = np.squeeze(hrv_indices.values)
        except:
            hrv_values = np.zeros((57))
        
        if len(hrv_values) != 57:
            hrv_values = np.zeros((57))
        sym_to_append = np.append(sym_to_append, hrv_values)
        #bio_df, bio_info = nk.bio_process(ecg=signal, sampling_rate=f500)        
        #results = nk.bio_analyze(bio_df, sampling_rate=f500)
        #r_peaks = bio_df['ECG_R_Peaks'][bio_df['ECG_R_Peaks'] !=0].index.values.astype(int)
        #r_mags = bio_df['ECG_Clean'][r_peaks].values
        #plotting_ECG_points(bio_df['ECG_Clean'], r_peaks)       
 
        # Get sym features
        sym_values = np.array([])
        ecg_features_lite = dict()
        ecg_features_lite['times'] = dict()
        ecg_features_lite['times']['current_ibi'] = np.diff(r_peaks) / f500
        ecg_features_lite['mags'] = dict()
        ecg_features_lite['mags']['current_mags'] = r_mags[:-1]
        sym_features = ecg_sym.get_sym_features(ecg_features_lite)
        data_to_append = {}        
        cols = list(sym_features.keys())        
        for col in cols:
            chunk = sym_features[col]
            for i, value in enumerate(chunk):
                col_name = 'sym_' + col + '__' + str(i)
                data_to_append[col_name] = value
                sym_values = np.append(sym_values, value)

        if len(sym_values) != 86:
            sym_values = np.zeros((86))
        sym_to_append = np.append(sym_to_append, sym_values)

        # Get melspec
        mel_spec = librosa.feature.melspectrogram(signal, n_mels=64, n_fft=2048, hop_length=512, sr=f500, power=2)
        if len(mel_spec.ravel()) != 640:
            mel_spec = np.zeros((64, 10))        
        sym_to_append = np.append(sym_to_append, mel_spec.ravel())

    # Check for nan and +-inf        
    sym_to_append[np.isinf(sym_to_append)] = 0
    sym_to_append[np.isneginf(sym_to_append)] = 0
    sym_to_append[np.isnan(sym_to_append)] = 0        

    sym_to_append = np.append(sym_to_append, age)
    sym_to_append = np.append(sym_to_append, sex)
    return sym_to_append
# -------------------------------- [ ] -------------------------------- #

# -------------------------------- [ get_features_mfcc ] -------------------------------- #
def get_features_mfcc(header, recording, leads):
    # Extract age.
    age = get_age(header)
    if age is None:
        age = float('nan')

    # Extract sex. Encode as 0 for female, 1 for male, and NaN for other.
    sex = get_sex(header)
    if sex in ('Female', 'female', 'F', 'f'):
        sex = 0
    elif sex in ('Male', 'male', 'M', 'm'):
        sex = 1
    else:
        sex = float('nan')

    # Extract fs.
    fs = int(get_frequency(header))
    if fs is None:
        fs = 500    

    # Reorder/reselect leads in recordings.
    available_leads = get_leads(header)
    indices = list()
    for lead in leads:
        i = available_leads.index(lead)
        indices.append(i)
    recording = recording[indices, :].astype(np.float32)

    # Pre-process recordings.
    adc_gains = get_adc_gains(header, leads)
    baselines = get_baselines(header, leads)
    num_leads = len(leads)
    for i in range(num_leads):
        recording[i, :] = (recording[i, :] - baselines[i]) / adc_gains[i]

    #plot_ecg(recording[0, :])
    
    # Preprocess ECG
    #rms = np.zeros(num_leads, dtype=np.float32)    
    f500 = 500
    sym_to_append = np.array([])
    for i in range(num_leads):
        x = recording[i, :]        
        signal = resample(x, fs, f500)
        signal = ecg_preprocess(signal, f500)  # preprocessing
        signal = repeat_until_n(signal, 10 * f500)  # 10 sec segment
        signal[np.isnan(signal)] = 0        
        mel_spec = librosa.feature.melspectrogram(signal, n_mels=64, n_fft=2048, hop_length=512, sr=f500, power=2)
        sym_to_append = np.append(sym_to_append, mel_spec.ravel())
        
    sym_to_append = np.append(sym_to_append, age)
    sym_to_append = np.append(sym_to_append, sex)
    return sym_to_append
# -------------------------------- [ ] -------------------------------- #

# -------------------------------- [ get_features_neurokit ] -------------------------------- #
# Extract features from the header and recording.
def get_features_neurokit(header, recording, leads):
    # Extract age.
    age = get_age(header)
    if age is None:
        age = float('nan')

    # Extract sex. Encode as 0 for female, 1 for male, and NaN for other.
    sex = get_sex(header)
    if sex in ('Female', 'female', 'F', 'f'):
        sex = 0
    elif sex in ('Male', 'male', 'M', 'm'):
        sex = 1
    else:
        sex = float('nan')

    # Extract fs.
    fs = int(get_frequency(header))
    if fs is None:
        fs = 500    

    # Reorder/reselect leads in recordings.
    available_leads = get_leads(header)
    indices = list()
    for lead in leads:
        i = available_leads.index(lead)
        indices.append(i)
    recording = recording[indices, :].astype(np.float32)

    # Pre-process recordings.
    adc_gains = get_adc_gains(header, leads)
    baselines = get_baselines(header, leads)
    num_leads = len(leads)
    for i in range(num_leads):
        recording[i, :] = (recording[i, :] - baselines[i]) / adc_gains[i]

    #plot_ecg(recording[0, :])
    
    # Preprocess ECG
    #rms = np.zeros(num_leads, dtype=np.float32)
    f500 = 500
    sym_to_append = np.array([])
    for i in range(num_leads):        
        x = recording[i, :]        
        signal = resample(x, fs, f500)        
        signal = ecg_preprocess(signal, f500)  # preprocessing
        #plot_2ecg(x, signal)
        signal = repeat_until_n(signal, 15 * f500)  # 15 sec segment
        signal[np.isnan(signal)] = 0
        r_peaks = processing.xqrs_detect(sig=signal, fs=f500, verbose=False, learn=True)
        if len(r_peaks) < 6:            
            r_peaks = np.sort(np.random.choice(range(15 * f500 - 1), 15, replace=False))
        r_peaks = peakDetectCorrect(signal, f500, r_peaks)
        r_mags = signal[r_peaks]
        #plotting_ECG_points(signal, r_peaks)

        try:
            hrv_indices = nk.hrv(r_peaks, sampling_rate=f500, show=False)
            hrv_values = np.squeeze(hrv_indices.values)
        except:
            hrv_values = np.zeros((52))

        #bio_df, bio_info = nk.bio_process(ecg=signal, sampling_rate=f500)        
        #results = nk.bio_analyze(bio_df, sampling_rate=f500)
        #r_peaks = bio_df['ECG_R_Peaks'][bio_df['ECG_R_Peaks'] !=0].index.values.astype(int)
        #r_mags = bio_df['ECG_Clean'][r_peaks].values
        #plotting_ECG_points(bio_df['ECG_Clean'], r_peaks)       
 
        # Check for nan and +-inf
        sym_to_append = np.append(sym_to_append, hrv_values)
        sym_to_append[np.isinf(sym_to_append)] = 0
        sym_to_append[np.isneginf(sym_to_append)] = 0
        sym_to_append[np.isnan(sym_to_append)] = 0        
        # Get sym features
        ecg_features_lite = dict()
        ecg_features_lite['times'] = dict()
        ecg_features_lite['times']['current_ibi'] = np.diff(r_peaks) / f500
        ecg_features_lite['mags'] = dict()
        ecg_features_lite['mags']['current_mags'] = r_mags[:-1]
        sym_features = ecg_sym.get_sym_features(ecg_features_lite)
        data_to_append = {}        
        cols = list(sym_features.keys())        
        for col in cols:
            chunk = sym_features[col]
            for i, value in enumerate(chunk):
                col_name = 'sym_' + col + '__' + str(i)
                data_to_append[col_name] = value
                sym_to_append = np.append(sym_to_append, value)

    sym_to_append = np.append(sym_to_append, age)
    sym_to_append = np.append(sym_to_append, sex)
    return sym_to_append
# -------------------------------- [ ] -------------------------------- #

# -------------------------------- [ PlottingECGpy ] -------------------------------- #
def plotting_ECG_points(Signal,Points):
    """
    Plot ECG-important points on signal    

    Args:
        Signal: ecg signal in mV
        Points: points

    Returns:

    """    
    plt.figure(0)
    plt.plot(Signal, '-rD', markevery=list(Points))
    plt.plot(Signal)
    plt.ylabel('U,mV')
    plt.xlabel('samples')
    plt.title('ECG')
    plt.grid(True)
    plt.show()
# ------------------------------------ [ ] ---------------------------------- #

# -------------------------------- [ plot_2ecg ] -------------------------------- #
def plot_2ecg(array1,array2):
    plt.figure(1)
    plt.plot(array1, '-b')
    plt.plot(array2, '-r')
    plt.ylabel('U,mV')
    plt.xlabel('samples')
    plt.title('ECG')
    plt.grid(True)
    plt.show()
# ------------------------------------ [ ] ---------------------------------- #

# -------------------------------- [ plot_ecg ] -------------------------------- #
def plot_ecg(ecg):
    """
    Plot ECG signal

    Args:
        ecg: ecg signal in mV

    Returns:

    """
    plt.figure(0)
    plt.plot(ecg)
    plt.ylabel('U,mV')
    plt.xlabel('samples')
    plt.title('ECG')
    plt.grid(True)
    plt.show()
# -------------------------------- [ ] -------------------------------- #        

# -------------------------------- [ create_train_test_folders_all_class ] -------------------------------- #
def create_train_test_folders_all_class(pathes_to_raw, path_to_train, path_to_test, classes, amount_per_train=10, amount_per_test=10):
    """
    Select and copy files for Train 

    Args:        
        pathes_to_raw: list with pathes to raw data
        path_to_train: path to train folder
        path_to_test: path to test folder
        classes: list with classes
        amount_per_train: amount of train instances (-1 for full data set, without test data)
        amount_per_test: amount of test instances

    Returns:

    """
    for path_to_raw in pathes_to_raw:
        header_files, recording_files = find_challenge_files(path_to_raw)            
        for i in range(len(header_files)):            
            header_fn = ntpath.basename(header_files[i])
            recording_fn = ntpath.basename(recording_files[i])
            copyfile(header_files[i], path_to_train + header_fn)
            copyfile(recording_files[i], path_to_train + recording_fn)
# -------------------------------- [ ] -------------------------------- #

# -------------------------------- [ create_train_test_folders_any_class ] -------------------------------- #
def create_train_test_folders_any_class(pathes_to_raw, path_to_train, path_to_test, classes, amount_per_train=10, amount_per_test=10):
    """
    Select and copy files for Train and Test sets, if single class is presented

    Args:        
        pathes_to_raw: list with pathes to raw data
        path_to_train: path to train folder
        path_to_test: path to test folder
        classes: list with classes
        amount_per_train: amount of train instances (-1 for full data set, without test data)
        amount_per_test: amount of test instances

    Returns:

    """
    normal_class = '426783006'
    equivalent_classes = [['713427006', '59118001'], ['284470004', '63593006'], ['427172004', '17338001']]    
    # Create train and test
    if amount_per_train > 0:
        counter_train = dict()
        counter_test = dict()
        for single_class in classes:
            counter_train[single_class] = 0
            counter_test[single_class] = 0                                    
        for path_to_raw in pathes_to_raw:
            header_files, recording_files = find_challenge_files(path_to_raw)            
            for single_class in classes:
                for i in range(len(header_files)):
                    if counter_train[single_class] < amount_per_train:
                        header = load_header(header_files[i])
                        current_labels = get_labels(header)
                        try:
                            current_labels = list(map(int, current_labels))                            
                        except:
                            current_labels = list()
                        if single_class in current_labels:
                            header_fn = ntpath.basename(header_files[i])
                            recording_fn = ntpath.basename(recording_files[i])
                            copyfile(header_files[i], path_to_train + header_fn)
                            copyfile(recording_files[i], path_to_train + recording_fn)
                            counter_train[single_class] += 1
                    elif counter_test[single_class] <  amount_per_test:
                        header = load_header(header_files[i])
                        current_labels = get_labels(header)
                        try:
                            current_labels = list(map(int, current_labels))
                        except:
                            current_labels = list()
                        if single_class in current_labels:
                            header_fn = ntpath.basename(header_files[i])
                            recording_fn = ntpath.basename(recording_files[i])
                            copyfile(header_files[i], path_to_test + header_fn)
                            copyfile(recording_files[i], path_to_test + recording_fn)
                            counter_test[single_class] += 1
        print(counter_train)
        print(counter_test)
    # Create full train only
    else:
        for path_to_raw in pathes_to_raw:
            header_files, recording_files = find_challenge_files(path_to_raw)
            for single_class in classes:
                for i in range(len(header_files)):
                    header = load_header(header_files[i])
                    current_labels = get_labels(header)
                    current_labels = list(map(int, current_labels))
                    if single_class in current_labels:
                        header_fn = ntpath.basename(header_files[i])
                        recording_fn = ntpath.basename(recording_files[i])
                        copyfile(header_files[i], path_to_train + header_fn)
                        copyfile(recording_files[i], path_to_train + recording_fn)
# -------------------------------- [ ] -------------------------------- #

# -------------------------------- [ create_train_test_folders ] -------------------------------- #
def create_train_test_folders(pathes_to_raw, path_to_train, path_to_test, classes, amount_per_train=10, amount_per_test=10):
    """
    Select and copy files for Train and Test sets

    Args:        
        pathes_to_raw: list with pathes to raw data
        path_to_train: path to train folder
        path_to_test: path to test folder
        classes: list with classes
        amount_per_train: amount of train instances (-1 for full data set, without test data)
        amount_per_test: amount of test instances

    Returns:

    """
    normal_class = '426783006'
    equivalent_classes = [['713427006', '59118001'], ['284470004', '63593006'], ['427172004', '17338001']]    
    # Create train and test
    if amount_per_train > 0:
        counter_train = dict()
        counter_test = dict()
        for single_class in classes:
            counter_train[single_class] = 0
            counter_test[single_class] = 0                                    
        for path_to_raw in pathes_to_raw:
            header_files, recording_files = find_challenge_files(path_to_raw)
            for single_class in classes:
                for i in range(len(header_files)):
                    if counter_train[single_class] < amount_per_train:
                        header = load_header(header_files[i])
                        current_labels = get_labels(header)
                        current_labels = list(map(int, current_labels))
                        if single_class in current_labels and set(current_labels).issubset(classes):
                            header_fn = ntpath.basename(header_files[i])
                            recording_fn = ntpath.basename(recording_files[i])
                            copyfile(header_files[i], path_to_train + header_fn)
                            copyfile(recording_files[i], path_to_train + recording_fn)
                            counter_train[single_class] += 1
                    elif counter_test[single_class] <  amount_per_test:
                        header = load_header(header_files[i])
                        current_labels = get_labels(header)
                        current_labels = list(map(int, current_labels))
                        if single_class in current_labels and set(current_labels).issubset(classes):
                            header_fn = ntpath.basename(header_files[i])
                            recording_fn = ntpath.basename(recording_files[i])
                            copyfile(header_files[i], path_to_test + header_fn)
                            copyfile(recording_files[i], path_to_test + recording_fn)
                            counter_test[single_class] += 1
        print(counter_train)
        print(counter_test)
    # Create full train only
    else:
        for path_to_raw in pathes_to_raw:
            header_files, recording_files = find_challenge_files(path_to_raw)
            for single_class in classes:
                for i in range(len(header_files)):
                    header = load_header(header_files[i])
                    current_labels = get_labels(header)
                    current_labels = list(map(int, current_labels))
                    if single_class in current_labels and set(current_labels).issubset(classes):
                        header_fn = ntpath.basename(header_files[i])
                        recording_fn = ntpath.basename(recording_files[i])
                        copyfile(header_files[i], path_to_train + header_fn)
                        copyfile(recording_files[i], path_to_train + recording_fn)
# -------------------------------- [ ] -------------------------------- #

# -------------------------------- [ reduce_mem_usage ] -------------------------------- #
def reduce_mem_usage(df, verbose=True):
    """
    Reduce memory usage
    https://www.kaggle.com/frtgnn/simple-xgb-k-fold-validation-approach
    """
    numerics = ['int16', 'int32', 'int64', 'float16', 'float32', 'float64']
    start_mem = df.memory_usage().sum() / 1024**2    
    for col in df.columns:
        col_type = df[col].dtypes
        if col_type in numerics:
            c_min = df[col].min()
            c_max = df[col].max()
            if str(col_type)[:3] == 'int':
                if c_min > np.iinfo(np.int8).min and c_max < np.iinfo(np.int8).max:
                    df[col] = df[col].astype(np.int8)
                elif c_min > np.iinfo(np.int16).min and c_max < np.iinfo(np.int16).max:
                    df[col] = df[col].astype(np.int16)
                elif c_min > np.iinfo(np.int32).min and c_max < np.iinfo(np.int32).max:
                    df[col] = df[col].astype(np.int32)
                elif c_min > np.iinfo(np.int64).min and c_max < np.iinfo(np.int64).max:
                    df[col] = df[col].astype(np.int64)  
            else:
                if c_min > np.finfo(np.float16).min and c_max < np.finfo(np.float16).max:
                    df[col] = df[col].astype(np.float16)
                elif c_min > np.finfo(np.float32).min and c_max < np.finfo(np.float32).max:
                    df[col] = df[col].astype(np.float32)
                else:
                    df[col] = df[col].astype(np.float64)    
    end_mem = df.memory_usage().sum() / 1024**2
    if verbose: print('Mem. usage decreased to {:5.2f} Mb ({:.1f}% reduction)'.format(end_mem, 100 * (start_mem - end_mem) / start_mem))
    return df
# -------------------------------- [ ] -------------------------------- #

# -------------------------------- [ resample ] -------------------------------- #
def resample(Signal, currentFs, targetFs):
    """
    Read ECG from different formats. Supported: kolibri_raw, wfdb

    Args:
        Signal: input signal
        currentFs: current sampling frequency
        targetFs: target sampling frequency

    Returns:
        finalArray: resampled signal
    """ 
    Signal = Signal.reshape(-1, 1)
    Signal = np.squeeze(Signal)
    n = len(Signal)
    nfinal = np.round(np.multiply(n, np.divide(targetFs, currentFs)))  # length of final signal
    y = np.floor(np.log2(n))
    nextpow2 = int(np.power(2, y + 1))    
    bigArray = np.zeros((nextpow2), dtype=np.float)
    bigArray[:n] = Signal
    resArray = sig.resample(bigArray, int(np.round(len(bigArray) * (targetFs / currentFs))))
    finalArray = resArray[:int(nfinal)]
    return finalArray
# ------------------------------------- [ ] ------------------------------------ #

# -------------------------------- [ repeat_until_n ] -------------------------------- #
def repeat_until_n(signal, n):
    """
    Extend signal with its copies to length n. If len(signal) > n -> len(signal)=n

    Args:
        signal: vector of input data
        n: necessary length

    Returns:
        out_signal: output vector
    """
    ns = len(signal)
    if ns >= n:
        out_signal = signal[:n]
    else:
        rep_amount = int(np.ceil(n / ns))
        out_signal = np.tile(signal, rep_amount)        
        out_signal = out_signal[:n]
    return out_signal
# -------------------------------- [  ] -------------------------------- #

# -------------------------------- [ peakDetectCorrect ] -------------------------------- #
""" 
    Peak detection using XQRS and correction in sliding window
"""
def peakDetectCorrect(Signal, Fs, qrsPeaks):
    aperture = 2
    outRange = 0
    firstValue = 0
    lastValue = 0

    Signal = Signal.reshape(-1, 1)
    Signal = np.squeeze(Signal)
    
    if (qrsPeaks[0] - aperture < 0) and (qrsPeaks[-1] + aperture) >= len(Signal):
        outRange = 3
        firstValue = qrsPeaks[0]
        lastValue = qrsPeaks[-1]
        qrsPeaks = qrsPeaks[1:-1]
    elif (qrsPeaks[-1] + aperture) >= len(Signal):
        outRange = 1
        lastValue = qrsPeaks[-1]
        qrsPeaks = qrsPeaks[:-1]
    elif (qrsPeaks[0] - aperture) < 0:
        outRange = 2
        firstValue = qrsPeaks[0]
        qrsPeaks = qrsPeaks[1:]

    segmentedSignal = np.zeros(shape=(len(qrsPeaks), aperture+aperture+1))
    localIdxs = np.arange(-aperture, aperture+1, 1)
    for lIdx in localIdxs:
        segmentedSignal[:, lIdx + aperture] = Signal[qrsPeaks + lIdx]
    peakIdxs = np.argmax(segmentedSignal, axis=-1)
    qrsPeaks = qrsPeaks - aperture + peakIdxs

    if outRange == 1:
        qrsPeaks = np.append(qrsPeaks, lastValue)
    elif outRange == 2:
        qrsPeaks = np.insert(qrsPeaks, 0, firstValue)
    elif outRange == 3:
        qrsPeaks = np.insert(qrsPeaks, 0, firstValue)
        qrsPeaks = np.append(qrsPeaks, lastValue)

    return qrsPeaks
# -------------------------------- [ ] -------------------------------- #

# -------------------------------- [ ecg_preprocess ] -------------------------------- #
def ecg_preprocess(signal, fs):
    """
    ECG preprocess (filter, cutting)

    Args:
        signal: vector of input ECG
        fs: sampling frequency

    Returns:
        out_signal: output ECG vector
    """
    #signal = nk.signal_filter(signal, 
    #                        sampling_rate=fs, 
    #                        method="powerline",
    #                        powerline=50,
    #                    )
    #signal = nk.signal_filter(signal, 
    #                        sampling_rate=fs, 
    #                        method="powerline",
    #                        powerline=60,
    #                    )              
    signal = nk.signal_filter(signal, 
                            sampling_rate=fs, 
                            lowcut=0.5, 
                            highcut=45, 
                            method="butterworth",
                            order=2,
                            window_size="default"                            
                        )      

    dyn_range = signal.max() - signal.min() 
    signal = signal / dyn_range if dyn_range < 0.5 else signal

    bad_frame_options = dict()
    bad_frame_options['fs'] = fs
    bad_frame_options['fs_int'] = 94
    bad_frame_options['frame_samples'] = 160
    bad_frame_options['mean_mv'] = 0.05
    bad_frame_options['range_mv'] = 3.2767
    bad_frame_options['threshold_mv'] = 0.05

    bad_frames_map, percent_detection = detect_bad_frames(signal, bad_frame_options)            
    signal = delete_bad_frames(signal, bad_frames_map, bad_frame_options, percent_detection)    
    
    return signal
# -------------------------------- [  ] -------------------------------- #

# -------------------------------- [ detect_bad_frames ] -------------------------------- #
def detect_bad_frames(Signal, Options):
    # a=a.reshape(a.shape[0],1) <- rank 1 array to vector    
    FrameLength = Options['fs'] * Options['frame_samples'] / Options['fs_int']
    N = len(Signal)
    NQ = int(N / FrameLength)
    frames = np.arange(0, NQ, 1,dtype=int)
    MeanSig = np.zeros((NQ, 1))
    SigRange = np.zeros((NQ, 1))
    MeanSigFinal = np.zeros((NQ, 1))

    for i in frames:
        MeanSig[i], SigRange[i] = calcMeanRange(Signal, i, FrameLength)

    BadFramesMap3 = np.absolute(MeanSig) < Options['mean_mv']

    MeanSigArray1 = MeanSig[0:-1]
    MeanSigArray2 = MeanSig[1:]
    MeanSigResult = MeanSigArray1 - MeanSigArray2
    MeanSigFinal[1:] = MeanSigResult

    BadFramesMap1 = MeanSigFinal < Options['threshold_mv']
    BadFramesMap2 = SigRange < Options['range_mv']
    BadFramesMap = np.bitwise_and(np.bitwise_and(BadFramesMap1, BadFramesMap2), BadFramesMap3)
    PercentDetection = np.count_nonzero(BadFramesMap) / len(BadFramesMap)
    return BadFramesMap, PercentDetection


def calcMeanRange(Signal,i,FrameLength):
    Start = int((i) * FrameLength)
    Stop = int((i+1)*FrameLength)
    SigTemp = Signal[Start:Stop]
    MeanSig = np.mean(SigTemp)
    SigRange = np.amax(SigTemp) - np.amin(SigTemp)
    return MeanSig, SigRange

# ------------------------------------- [  ] ----------------------------------- #

# -------------------------------- [ delete_bad_frames ] -------------------------------- #
def delete_bad_frames(Signal,BadFramesMap,Options,PercentDetection):

    if 0 < PercentDetection < 1:
        FrameLength = Options['fs'] * Options['frame_samples'] / Options['fs_int']
        oneFrame = np.zeros((int(FrameLength), 1), dtype=float)
        for i, val in enumerate(BadFramesMap):
            if BadFramesMap[i]:
                Start = int((i) * FrameLength)
                Stop = int((i + 1) * FrameLength)
                oneFrame = Signal[Start:Stop]
                if 'SignalNew' in locals():
                    SignalNew = np.concatenate((SignalNew, oneFrame), axis=0)
                else:
                    SignalNew = oneFrame
    else:
        SignalNew = Signal

    return SignalNew
# ------------------------------------ [ ] ------------------------------------- #

# -------------------------------- [ crop_or_pad ] -------------------------------- #
def crop_or_pad(y, length, is_train=True, start=None):
    if len(y) < length:
        #y = np.concatenate([y, np.zeros(length - len(y))])
        
        n_repeats = length // len(y)
        epsilon = length % len(y)
        
        y = np.concatenate([y]*n_repeats + [y[:epsilon]])
        
    elif len(y) > length:
        if not is_train:
            start = start or 0
        else:
            start = start or np.random.randint(len(y) - length)

        y = y[start:start + length]

    return y
# ------------------------------------ [ ] ------------------------------------- #

# -------------------------------- [ mono_to_color ] -------------------------------- #
def mono_to_color(X, eps=1e-6, mean=None, std=None):
    mean = mean or X.mean()
    std = std or X.std()
    X = (X - mean) / (std + eps)
    
    _min, _max = X.min(), X.max()

    if (_max - _min) > eps:
        V = np.clip(X, _min, _max)
        V = 255 * (V - _min) / (_max - _min)
        V = V.astype(np.uint8)
    else:
        V = np.zeros_like(X, dtype=np.uint8)

    return V
# ------------------------------------ [ ] ------------------------------------- #